/*************************************************************************** 
 * Copyright (c) Amir Hermelin 1999
 *
 * This software may be freely used and distributed provided:
 * 1) No charge is made for this software.
 * 2) This and the following copyright statements are retained.
 * 3) User accepts that this software is provided "AS IS"
 * with NO WARRANTIES WHATSOEVER, including but not limited to
 * merchantability, fitness for a particular purpose, and non-
 * infringement; and accepts SOLE LIABILITY for any consequences
 * whatsoever arising directly or indirectly from any use or
 * misuse of this software or related documentation.
 * 4) User agrees that the author is under no obligation to
 * maintain, update, modify, support or enhance this software
 * or related documentation.
 ***************************************************************************/

/*************************************************************************** 
 * Version History:
 *   0.9:  tputs command available, outputs to stdout
 *   0.92: tputs receives "puts" parameters and uses the "puts" command
 *          to print the string. Channels can be specified. A few minor
 *          bugs corrected
 *   1.0:  Added tputs_getStr
 *   1.09: Added switches, corrected for tighter error detection. Added
 *          rollbacks in case of an unfound code.
 *   1.1:  Added parameter passing
 *   1.2:  Added tputs_getNum, tputs_getFlag, and tputs_goto
 ***************************************************************************/
 
#include <stdlib.h> 
#include <stdio.h>
#include <string.h>

#include <tcl.h>
#include <termcap.h>


#ifndef TRUE
   #define TRUE 1
#endif
#ifndef FALSE
   #define FALSE 0
#endif
#ifndef OOPS
   #define OOPS -1
#endif


#define TPUTS_WITH_PARAMS
/* #define TPUTS_WITH_GOTO */


/***********************
 * Globals
 ***********************/

static const char* copyright  = "Copyright (c) Amir Hermelin 1999";
static       char* version    = "1.2";

#ifdef TPUTS_WITH_PARAMS
static const char *paramDelim = ",";
#endif

enum {
   TPUTS,
   TPUTS_GETSTR,
   TPUTS_GETNUM,
   TPUTS_GETFLAG,
   TPUTS_GOTO,
};

/****************
 * Defines
 ****************/   
 
/* BUFSIZ defined in stdio.h */
#define TPUTS_BUFSIZE_TOTALCOMMANDSIZE BUFSIZ+16+64
#define TPUTS_BUFSIZE_TERMCAPSEQUENCE  64
#define TPUTS_MAX_NUM_PARAMS 8

#define TPUTS_DEF_DELIMBEGIN '<'
#define TPUTS_DEF_DELIMEND   '>'


/***********************
 * Forward declerations
 ***********************/

int Tputs_cmd(ClientData clientData, Tcl_Interp *interp,
              int objc, Tcl_Obj *CONST objv[]); 

int Tputs_getstr_cmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[]);

int Tputs_getnum_cmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[]);

int Tputs_getflag_cmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[]);

int Tputs_goto_cmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[]);

static char* Tputs_common(ClientData clientData, Tcl_Interp *interp,
                          int objc, Tcl_Obj *CONST objv[],
                          int mode);

/*****************************************
 * Tputs_Init
 *
 * This is the shared library entry point.
 * It also registers the package using 
 *   pkg_mkIndex
 *****************************************/
 
int Tputs_Init(Tcl_Interp *interp)
{
  static int bIsFirstTime = TRUE;

  if (bIsFirstTime) {
    copyright = copyright; /* discard warning */
    Tcl_CreateObjCommand(interp, "tputs",         Tputs_cmd,         NULL, NULL);
    Tcl_CreateObjCommand(interp, "tputs_getStr",  Tputs_getstr_cmd,  NULL, NULL);
    Tcl_CreateObjCommand(interp, "tputs_getNum",  Tputs_getnum_cmd,  NULL, NULL);
    Tcl_CreateObjCommand(interp, "tputs_getFlag", Tputs_getflag_cmd, NULL, NULL);
   #ifdef TPUTS_WITH_GOTO
    Tcl_CreateObjCommand(interp, "tputs_goto",    Tputs_goto_cmd,    NULL, NULL);
   #endif
    
    Tcl_PkgProvide(interp, "tputs", version);
    bIsFirstTime = FALSE;
  }
  
  return TCL_OK;
}


/*********************************************
 * Tputs_cmd
 *
 * This is the tcl "tputs" command entry point.
 *********************************************/
int Tputs_cmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[])
{
   char* ret = Tputs_common(clientData, interp, objc, objv, TPUTS);
   if (ret == NULL)
      return TCL_ERROR;
   if (*ret == 0)
      return TCL_OK;
   return Tcl_Eval(interp, ret);
}

/*********************************************
 * Tputs_goto_cmd
 *
 * This is the tcl "tputs_goto" command entry point.
 *********************************************/
int Tputs_goto_cmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[])
{
   char* ret = Tputs_common(clientData, interp, objc, objv, TPUTS_GOTO);
   if (ret == NULL)
      return TCL_ERROR;
   if (*ret == 0)
      return TCL_OK;
   return Tcl_Eval(interp, ret);
}


/*********************************************
 * Tputs_getstr_cmd
 *
 * This is the tcl "tputs_getCode" command entry point.
 *********************************************/
int Tputs_getstr_cmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[])
{
   char* ret = Tputs_common(clientData, interp, objc, objv, TPUTS_GETSTR);
   if (ret == NULL)
      return TCL_ERROR;
   if (*ret != 0)
      Tcl_SetResult(interp, ret, TCL_STATIC);
   return TCL_OK;
}

/*********************************************
 * Tputs_getnum_cmd
 *
 * This is the tcl "tputs_getNum" command entry point.
 *********************************************/
int Tputs_getnum_cmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[])
{
   char* ret = Tputs_common(clientData, interp, objc, objv, TPUTS_GETNUM);
   if (ret == NULL)
      return TCL_ERROR;
   if (*ret != 0)
      Tcl_SetResult(interp, ret, TCL_STATIC);
   return TCL_OK;
}

/*********************************************
 * Tputs_getflag_cmd
 *
 * This is the tcl "tputs_getFlag" command entry point.
 *********************************************/
int Tputs_getflag_cmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[])
{
   char* ret = Tputs_common(clientData, interp, objc, objv, TPUTS_GETFLAG);
   if (ret == NULL)
      return TCL_ERROR;
   if (*ret != 0)
      Tcl_SetResult(interp, ret, TCL_STATIC);
   return TCL_OK;
}



/************************************************
 * Tputs_common
 *
 * The main module and functionality of "tputs".
 * Returns the formatted string; adds extra
 *  info to the "puts" command if mode is PUTSF
 ************************************************/
static char* Tputs_common(ClientData clientData, Tcl_Interp *interp,
                          int objc, Tcl_Obj *CONST objv[],
                          int mode)
{
   static int  bFormattingOn = TRUE;   /* Is formatting enabled         */
   static char termDataBuf[1024];      /* The terminal data buffer      */
   static char errBuf[80];             /* The returned error buffer     */
   static char delimBegin = TPUTS_DEF_DELIMBEGIN; /* The control-code begin char   */
   static char delimEnd   = TPUTS_DEF_DELIMEND;   /* The control-code end char     */
   const  char delimDisable = '\\';    /* The control-code disable char */
   
   char* termNamePtr;                  /* The terminal name             */
   char* additionalParam;              /* Additional params to pass to 
                                           the "puts" command           */
   static   char outBuf[TPUTS_BUFSIZE_TOTALCOMMANDSIZE];
                                       /* The command-output buffer     */
            char* outPtr;              /* Pointer to the next output    */
   register char* inPtr;               /* Pointer to the next input     */
   
   static int bIsFirstTime = 1;        /* Used for one-time initialization */
   
   int bNoFormattingChar = FALSE;/* Set when the '\' char was encountered  */
   int bOptNonewline     = FALSE;/* Set when the -nonewline switch exists  */
   int bOptNoformat      = FALSE;
   int bOptExact         = FALSE;
   int iFirstPutsParam = 1;      /* The first parameter to "puts"          */
   int i, error;
   int bPolicyFormat     = FALSE;
   int bPolicyExact      = FALSE;
   int bAddPutsPadding   = (mode == TPUTS || mode == TPUTS_GOTO);
/*   int bPolicyNumericParams = bEnableNumericCapab;*/
   int res = OOPS;               /* For the numeric and boolean capabilities */
   
   clientData = clientData; /* supress warning */
   
   /* Init */
   outPtr = outBuf;
   *outPtr = 0;
   
   /**********************************************
    * Initialize first time only.
    * Load the database, check for errors.
    **********************************************/
   if (bIsFirstTime) {
      bIsFirstTime = 0;
      if ( (termNamePtr = Tcl_GetVar2(interp, "env", "TERM", 0)) == NULL )
{
         Tcl_SetResult(interp, "Can't read TERM environment variable", TCL_STATIC);
         bFormattingOn = FALSE;         
         return NULL;
      }
      /* Load the terminal capabilities database */
      switch (tgetent(termDataBuf, termNamePtr)) {
      case -1:
         Tcl_SetResult(interp, "Can't open termcap database", TCL_STATIC);
         bFormattingOn = FALSE;         
         return NULL;
      case 0:
         sprintf(errBuf, "Can't find database for %s", termNamePtr);
         Tcl_SetResult(interp, errBuf, TCL_STATIC);
         bFormattingOn = FALSE;         
         return NULL;
      default:
      }
   }
   
   
   /**********************************************
    * Check for parameters
    * Assign formatting policies
    **********************************************/
   while (iFirstPutsParam < objc) {
      additionalParam = Tcl_GetStringFromObj(objv[iFirstPutsParam], NULL);
      if (additionalParam == NULL) {
         Tcl_SetResult(interp, "Unexpected error", TCL_STATIC);
         return NULL;
      }
      /* traverse switch options */
      if ( ! strcmp(additionalParam, "-formatoff") ) {
         bFormattingOn = FALSE;
         
      } else if ( ! strcmp(additionalParam, "-formaton") ) {
         bFormattingOn = TRUE;
         
      } else if ( ! strcmp(additionalParam, "-nonewline") ) {
         bOptNonewline = TRUE;
         
      } else if ( ! strcmp(additionalParam, "-exact") ) {
         bOptExact = TRUE;
         
      } else if ( ! strcmp(additionalParam, "-noformat") ) {
         bOptNoformat = TRUE;
         
      } else if ( ! strcmp(additionalParam, "-controlchars") ) {
         int length; /* must be two on return value */
         
         if (++iFirstPutsParam == objc) {
            Tcl_SetResult(interp, "bad option \"-controlchars\": missing input", TCL_STATIC);
            return NULL;
         }
         additionalParam = Tcl_GetStringFromObj(objv[iFirstPutsParam], &length);
         if (additionalParam == NULL) {
            Tcl_SetResult(interp, "Unexpected error", TCL_STATIC);
            return NULL;
         }
         if (length > 2) {
            sprintf(errBuf, "bad option \"-controlchars\": too many chars");
         } else if (length < 2) {
            sprintf(errBuf, "bad option \"-controlchars\": too few chars");
         }
         if (length != 2) {
            Tcl_SetResult(interp, errBuf, TCL_STATIC);
            return NULL;
         }
         /* set delimiters */
         delimBegin = additionalParam[0];
         delimEnd   = additionalParam[1];
         
      } else if ( ! strcmp(additionalParam, "-controlcharsdef") ) {
         delimBegin = TPUTS_DEF_DELIMBEGIN;
         delimEnd   = TPUTS_DEF_DELIMEND;
         
      } else if ( ! strcmp(additionalParam, "--") ) {
         /* end of switches */
         iFirstPutsParam++;
         break;
         
      } else {
         break;
         
      }
      iFirstPutsParam++;
   }
   
   /* policies */
   bPolicyFormat = ( (!bOptNoformat) && bFormattingOn ) ;
   
   /* Check for wrong number of params - this is necessary so 
       that "tputs" returns an error and not "puts" */
   if (objc > 3 + iFirstPutsParam || objc < 1 + iFirstPutsParam) {
      if (iFirstPutsParam != 1  && iFirstPutsParam == objc) {
         /* only switches were entered - command is valid */
         /* do nothing */
         *outBuf = 0;
         return outBuf;
      }
      if (mode == TPUTS)
         sprintf(errBuf, "?switches? ?channelId? string");
      else
         sprintf(errBuf, "?switches? string");
      Tcl_WrongNumArgs(interp, 1, objv, errBuf);
      return NULL;
   } 
   
   /**********************************************
    * Prepare the output buffer, to be passed
    *  to the "puts" command. Pass additional
    *  parameters as well.
    **********************************************/
   if (bAddPutsPadding) {
      /* add extra formatting to pass to the puts command */
      outPtr += sprintf(outBuf, "puts ");
      if (bOptNonewline) {
         outPtr += sprintf(outPtr, "-nonewline ");
      }
      for (i = iFirstPutsParam; i < objc - 1; i++) {
         additionalParam = Tcl_GetStringFromObj(objv[i], NULL);
         if (additionalParam == NULL) {
            Tcl_SetResult(interp, "Unexpected error", TCL_STATIC);
            return NULL;
         }
         outPtr += sprintf(outPtr, "%s ", additionalParam);
      }
      *outPtr++ = '{';
   }
   inPtr = Tcl_GetStringFromObj(objv[objc - 1], NULL);
   if (inPtr == NULL) {
      Tcl_SetResult(interp, "Unexpected error", TCL_STATIC);
      return NULL;
   }
   if (strlen(inPtr) > TPUTS_BUFSIZE_TOTALCOMMANDSIZE - 200) {
      /* Last parameter string is too big */
      sprintf(errBuf, "String too big");
      Tcl_SetResult(interp, errBuf, TCL_STATIC);
      return NULL;
   }
   
   /**********************************************
    * The formatting loop
    * 1) Go over the input
    **********************************************/
   while (*inPtr) {
      bPolicyExact = ( (!bNoFormattingChar) && (!bOptExact) );
      if (*inPtr == delimBegin && bPolicyExact) {
         /**********************************************
          * 2) Received beginning of code
          **********************************************/
         char* idPtr ;
         char  idBuf[TPUTS_BUFSIZE_TERMCAPSEQUENCE];
         char  idParamBuf[TPUTS_BUFSIZE_TERMCAPSEQUENCE];
         
         idPtr = idBuf;
         while (*(++inPtr) && *inPtr != delimEnd)
            *idPtr++ = *inPtr;
         *idPtr = 0;
         /*************************************************
          * 3) Finished reading the code; resists in idBuf
          *************************************************/
         if (!(*inPtr)) {
            /* Error - end of input was met before closing
             *   delimiter. 
             * Rollback. */
            outPtr += sprintf(outPtr, "%c%s", delimBegin, idBuf);
            break; /* break out of while loop */
         } else {
            /* not premature end-of-input */
            if (bPolicyFormat) {
               /****************************************************
                * 4) Policy dictates to do formatting.
                * Check for parameters to termcap code.
                * In the format of <code,1,2,3>
                ****************************************************/
                char termcapCode[80];
                char *termcapCodePtr = termcapCode;
                int  aParams[TPUTS_MAX_NUM_PARAMS];
                int  nParams = 0;
                char* strret;
            #ifdef TPUTS_WITH_PARAMS
                char *token;
            #endif

                strcpy(idParamBuf, idBuf);
            #ifdef TPUTS_WITH_PARAMS
                token = strtok(idParamBuf, paramDelim);
            
                token = strtok(NULL, paramDelim);
                /* Get parameters */
                while (token != NULL && nParams < TPUTS_MAX_NUM_PARAMS) {
                    /* Determine whether a , or ; was entered 
                    char ch = idBuf[token - 1 - idParamBuf];
                    switch (ch) {
                    case ',':
                        aParams[nParams++] = atoi(token);
                    case ';':
                        aParams[nParams++] = atoi(token);
                    }
                    */
                    aParams[nParams++] = atoi(token);
                    token = strtok(NULL, paramDelim);
                }
            #else
               nParams = 0;
            #endif
           
                /* Get the termcap code */
                error = FALSE;
                switch (mode) {
                case TPUTS_GOTO:
                   /* Goto mode */
                   if (nParams != 2)
                      strret = NULL;
                   else {
                      strret = tgoto(idParamBuf, aParams[0], aParams[1]);
                      if (strret != NULL)
                         outPtr += sprintf(outPtr, "%s", strret);  
                   }
                   if (strret == NULL)
                      error = TRUE;
                   break;
                case TPUTS:
                case TPUTS_GETSTR:
                   /* String mode */
                   strret= tgetstr(idParamBuf, &termcapCodePtr);
                   
                   if (strret != NULL) {
                      /* Test parameters */
                      if (nParams == 0 )
                          strret = termcapCode; 
                      else {
                         strret = tparm(termcapCode,
                                        aParams[0],
                                        aParams[1],
                                        aParams[2],
                                        aParams[3],
                                        aParams[4],
                                        aParams[5],
                                        aParams[6],
                                        aParams[7] );
                      }
                      outPtr += sprintf(outPtr, "%s", strret);
                      break;
                   }
                   /* Fall through */
                case TPUTS_GETNUM:
                   /* Numeric mode */
                   if ((res = tgetnum (idParamBuf)) == OOPS)
                     error = TRUE;
                   else
                     outPtr += sprintf(outPtr, "%d", res);
                   break;
                case TPUTS_GETFLAG:
                   /* Flag mode */
                   if ((res = tgetflag (idParamBuf)) == OOPS)
                     error = TRUE;
                   else
                     outPtr += sprintf(outPtr, "%d", res);
                   break;
                }
                if (error) {
                     /* Error - termcap code doesn't exist */
                     /* rollback */
                     outPtr += sprintf(outPtr, "%c%s%c", delimBegin, idBuf, delimEnd);
                }
            }     /* end step 4 */
         }     /* end step 3 */
      }  /* end step 2 */
      else {
         /* no formatting code */
         *outPtr++ = *inPtr;
         bNoFormattingChar = (*inPtr == delimDisable);
      }
      inPtr++;
   }  /* end step 1 */

   /**********************************************
    * Finish
    **********************************************/
   /* finish the string to return */
   if (bAddPutsPadding)
      *outPtr++ = '}';
   *outPtr = 0;
   
   return outBuf;
}


